from typing import IO
import pydantic
from .cfg import FileConfig, load_config
from .domain import AstUnionNode, Config, Emitter
from .mkast import get_emitter
import argparse as ap
import pathlib
import yaml
import sys
from .__version__ import __version__

# forward declare with strings
AstNodeModel = pydantic.RootModel[dict[str, "AstNodeModel | str"] | None]
AstUnionNodeModel = pydantic.RootModel[dict[str, AstNodeModel]]
# resolve forward refs
AstUnionNodeModel.model_rebuild()


def make_parser() -> ap.ArgumentParser:
    p = ap.ArgumentParser('mkast', description="""
Generate an AST data structure from a language-agnostic description expressed in YAML.

Options take precedence over values in the config file, if provided.
""", formatter_class=ap.RawTextHelpFormatter)
    _ = p.add_argument('-c', '--config', help='config file')
    _ = p.add_argument('input', nargs='?', type=str, help='input file (default stdin)')
    _ = p.add_argument('-t', '--target', help='target language')
    _ = p.add_argument('--known-type', nargs='*', type=set, help='known type', dest='known-types')
    _ = p.add_argument('--common-prop', nargs='*', type=prop, help="common property: 'name:type'", dest='common-props')
    _ = p.add_argument('--root', help='root node')
    _ = p.add_argument('--namespace', help='namespace')
    _ = p.add_argument('--assert', help='Expands to an assertion statement. $1 is replaced by the boolean expression to assert')
    _ = p.add_argument('--union', help='Expands to the declaration of an union node. $1 is replaced by the name of the node.')
    _ = p.add_argument('--product', help='Expands to the declaration of an product node. $1 is replaced by the name of the node.')
    _ = p.add_argument('--imports', '--import', help='imports to add at the top of the file', nargs='*')
    _ = p.add_argument('--dump', action='store_true', help='Dump config and input and exit')
    _ = p.add_argument('--version', action='store_true', help='Show version exit')
    return p

def prop(input: str) -> tuple[str, str]:
    s = input.split(':', 1)
    if len(s) < 2:
        raise ap.ArgumentTypeError(f"invalid property format: '{input}': missing separator")
    return s[0], s[1]


def config_from_args(args: ap.Namespace) -> Config:
    return Config.model_validate({k: v for k, v in vars(args).items() if v is not None})


def read_config(filename: pathlib.Path) -> Config:
    visited = set()
    cfg = None
    cfgF = None
    while not cfg or not cfgF or (cfgF.extends and (filename := (filename.parent / cfgF.extends).resolve()) not in visited):
        visited.add(filename)
        cfgF = read_config_file(filename)
        cfg = cfg + cfgF if cfg else cfgF
    return cfg

def read_config_file(filename: pathlib.Path) -> FileConfig:
    try:
        with open(filename) as f:
            return load_config(f)
    except OSError as e:
        print(f"warning: couldn't read config file '{filename}': {e}", file=sys.stderr)
    except pydantic.ValidationError as e:
        print(f"warning: invalid config '{filename}': {e}", file=sys.stderr)
    return FileConfig()

def load_input(stream: IO) -> tuple[Config | None, AstUnionNode]:
    docs = tuple(yaml.safe_load_all(stream))
    match len(docs):
        case 2:
            return Config.model_validate(docs[0]), AstUnionNodeModel(docs[1]).model_dump()
        case 1:
            return None, AstUnionNodeModel(docs[0]).model_dump()
        case i:
            raise ValueError(f"expected 1 or 2 documents, not {i}")


def parse_args() -> tuple[Config, AstUnionNode, Emitter]:
    p = make_parser()
    args = p.parse_args()
    if args.version:
        print(f'mkast {__version__}')
        exit()
    cfg = config_from_args(args)
    if args.config:
        cfg += read_config(pathlib.Path(args.config))
    try:
        input_cfg, input = load_input(ap.FileType()(str(cfg.input) if cfg.input else '-'))
    except pydantic.ValidationError as e:
        p.error(f"invalid input{f" '{cfg.input}'" if cfg.input else ''}: {e}")
    except ValueError as e:
        p.error(' '.join(e.args))
    if input_cfg:
        cfg += input_cfg
    if args.dump:
        print('CONFIG:', cfg)
        print('INPUT:', input)
        exit()
    # get the emitter early so we can error cleanly if it doesn't exist
    return cfg, input, get_emitter(cfg) or p.error(f"unknown target '{cfg.target}'")



from inspect import isclass
import pathlib
import sys
import yaml
import pydantic
from typing import IO, Literal, cast
from collections.abc import Callable, Sequence, Set, Mapping

ModifierKey = Literal['', '?', '+', '*']


class Modifier(pydantic.BaseModel):
    type: str = '$1'
    must: str | None = None
    none_when: str | None = None
    unwrap: str = '$1'


class Config(pydantic.BaseModel):
    input: pydantic.FilePath | None = None
    target: str = 'agnostic'
    known_types: Set[str] = set()
    common_props: Mapping[str, str] = {}
    root: str | None = None
    namespace: str | None = None
    assert_: str | None = pydantic.Field(alias='assert', default=None)
    union: str | None = None
    product: str | None = None
    imports: Sequence[str] = []
    modifiers: dict[ModifierKey, Modifier] = {}

    def merge(self, new: 'Config'):
        return _deep_merge(self, new)
    
    def __add__(self, new: 'Config'):
        return self.merge(new)


def _normalize[Abstract, Concrete](obj: Abstract, cls: Callable[[Abstract], Concrete]) -> Concrete:
    return obj if isclass(cls) and isinstance(obj, cast(type[Concrete], cls)) else cls(obj)


def _deep_merge[T](base: T, new: T) -> T:
    """ base overrides new
    merge algorithm:
    - sub-models: merge fields recursively
    - sequences/lists: concatenate
    - sets: union
    - dicts: union, recursively merge intersecting values
    - everything else: replace base by new unless new is None
    """
    if new is None:
        return base
    if isinstance(base, pydantic.BaseModel):
        assert isinstance(new, pydantic.BaseModel)
        return base.model_validate(_deep_merge(base.model_dump(exclude_unset=True), new.model_dump(exclude_unset=True)))
    if isinstance(base, Mapping):
        assert isinstance(new, Mapping)
        return cast(T, {k: _deep_merge(base.get(k), new.get(k)) for k in _normalize(base, set) | _normalize(new, set)})
    if isinstance(base, Set):
        assert isinstance(new, Set)
        return base | new
    if isinstance(base, Sequence) and not isinstance(base, (str, bytes)):
        assert isinstance(new, Sequence) and not isinstance(new, (str, bytes))
        return cast(T, _normalize(base, list) + _normalize(new, list))
    return new

class FileConfig(Config):
    extends: str | None = None


def load_config(stream: IO) -> FileConfig:
    return FileConfig.model_validate(yaml.safe_load(stream))



from ..cfg import Modifier, ModifierKey
from ..domain import Config, Emitter, NodeInfo, NodeKind, camelize, pascalize, get_dont_touch_me
from ..util import println, csl, cslq, remove_prefix, sub

from collections.abc import Callable, Iterable, Mapping
from itertools import chain
import re

Keywords = {
    'abstract', 'as', 'base', 'bool', 'break', 'byte', 'case', 'catch', 'char', 'checked', 'class', 'const', 'continue',
    'decimal', 'default', 'delegate', 'do', 'double', 'else', 'enum', 'event', 'explicit', 'extern', 'false', 'finally', 'fixed',
    'float', 'for', 'foreach', 'goto', 'if', 'implicit', 'in', 'int', 'interface', 'internal', 'is', 'lock', 'long', 'namespace',
    'new', 'null', 'object', 'operator', 'out', 'override', 'params', 'private', 'protected', 'public', 'readonly', 'ref',
    'return', 'sbyte', 'sealed', 'short', 'sizeof', 'stackalloc', 'static', 'string', 'struct', 'switch', 'this', 'throw', 'true',
    'try', 'typeof', 'uint', 'ulong', 'unchecked', 'unsafe', 'ushort', 'using', 'virtual', 'void', 'volatile', 'while'
}


class CSharpEmitter(Emitter):
    def __init__(self, cfg: Config):
        super().__init__(cfg)

        self.usings = set(cfg.imports)

        if cfg.assert_:
            self.assert_ = cfg.assert_
        else:
            self.usings.add('System.Diagnostics')
            self.assert_ = 'Debug.Assert($1);'

        # $1.All($1 => $2)
        # $2: inner
        # $1: name
        self.modifiers: dict[ModifierKey, Modifier] = {
            '': cfg.modifiers.get('', Modifier()),
            '?': cfg.modifiers.get('?', Modifier(type='$1?', none_when='$1 is null')),
            '+': cfg.modifiers.get('+', Modifier(type='IReadOnlyList<$1>', must='$1.Count > 0', unwrap='$1.All($1 => $2)')),
            '*': cfg.modifiers.get('*', Modifier(type='IReadOnlyList<$1>', unwrap='$1.All($1 => $2)')),
        }

        self.node_kinds = {
            NodeKind.Product: cfg.product or 'public sealed class $1',
            NodeKind.Union: cfg.union or 'public interface $1'
        }

    def intro(self):
        if self.usings:
            for using in sorted(self.usings):
                print(f'using {using};')
            print()

        if self.cfg.namespace:
            print(f'namespace {self.cfg.namespace};')
            print()

        if not self.cfg.root:
            return 0

        print(sub(self.node_kinds[NodeKind.Union], 1, pascalize(self.cfg.root)))
        print('{')
        for p in self.cfg.common_props.items():
            self.put_prop(1, self.cfg.root, *p)
        return 1

    def enter_node(self,
                   lvl: int,
                   parent: NodeInfo | None,
                   node: NodeInfo,
                   implements: Mapping[str, NodeKind],
                   props: Mapping[str, str]):
        if reserved_props := props & self.cfg.common_props.keys():
            raise ValueError(f"reserved property names in '{node.name}': {cslq(reserved_props)}")

        props = dict(chain(self.cfg.common_props.items(), props.items()))

        need_explicit_constructor = any(map(self.requires_validation, props.values()))

        nk = self.node_kinds[node.kind]
        is_record = re.search(r'\brecord\b', nk) is not None
        println(lvl, sub(nk, 1, pascalize(node.name)), end='')

        # primary constructor arguments
        if node.kind is NodeKind.Product and props and not need_explicit_constructor:
            print(f'({csl(map(self.argument(pascalize if is_record else camel_ident), props.items()))})', end='')

        # base types list
        print(base_type_list((parent.name,) + tuple(implements.keys())
                             if parent and parent.kind is NodeKind.Union else
                             implements), end='')

        print()
        println(lvl, '{')

        lvl += 1
        if (not is_record or need_explicit_constructor) and node.kind is NodeKind.Product and props:
            if need_explicit_constructor:
                println(lvl, f'public {pascalize(node.name)}({csl(map(self.argument(camel_ident), props.items()))})')
                println(lvl, '{')
                for p in props.items():
                    self.put_assignment(lvl + 1, *p)
                println(lvl, '}')
                for p in props.items():
                    self.put_prop(lvl, node.name, *p, 'public')
            else:
                # primary constructor initializers
                for p in props.items():
                    self.put_prop(lvl, node.name, *p, 'public', True)

    def exit_node(self, lvl: int):
        println(lvl, '}')

    def conclusion(self):
        if self.cfg.root:
            print('}')

    def argument(self, name_transformer: Callable[[str], str]):
        return lambda prop: f'{self.real_type(prop[1])} {name_transformer(prop[0])}'

    def put_assignment(self, lvl: int, name: str, type: str):
        if val_expr := self.validation_expr(camel_ident(name), type):
            println(lvl, sub(self.assert_, 1, val_expr))
        println(lvl, f'{pascalize(name)} = {camel_ident(name)};')

    def put_prop(self, lvl: int, owner: str, name: str, type: str, access: str = '', put_init: bool = False):
        access = access + ' ' if access else ''
        init = ' = ' + camel_ident(name) + ';' if put_init else ''
        println(lvl, f'{access}{self.real_type(remove_prefix(owner + ".", type))} {pascalize(name)} {{ get; }}{init}')

    def real_type(self, type: str) -> str:
        if type[-1] not in self.modifiers:
            return pascalize(type)
        m = self.modifiers[type[-1]]
        return sub(m.type, 1, self.real_type(type[:-1]))

    def validation_expr(self, name: str, type: str):
        if type[-1] not in self.modifiers:
            return ''
        m = self.modifiers[type[-1]]

        # {none_when} || {must} && {unwrap}
        none_when = sub(m.none_when or '', 1, name)
        must = sub(m.must or '', 1, name)

        if '$2' in m.unwrap and (inner := self.validation_expr(name, type[:-1])):
            unwrap = sub(sub(m.unwrap, 1, name), 2, inner)
        else:
            name = sub(m.unwrap, 1, name)
            unwrap = self.validation_expr(name, type[:-1])

        r = ' && '.join(filter(None, (must, unwrap)))
        return ' || '.join(filter(None, (none_when, r)))

    def requires_validation(self, type: str):
        validating_modifiers = {k for k, v in self.modifiers.items() if v.must}
        return any(c in validating_modifiers for c in type)


def base_type_list(bases: Iterable[str]):
    return ' : ' + csl(map(pascalize, bases)) if bases else ''


def camel_ident(name: str):
    if s := get_dont_touch_me(name):
        return s
    name = camelize(name)
    return '@' + name if name in Keywords else name



# ruff: noqa: E402,F401

from .__version__ import __version__

# load input files and  their config
from .cli import load_input
# load configs, config types
from .cfg import load_config, Config, Modifier, ModifierKey
# generate the ast, manage custom emitters
from .mkast import generate_ast, register_emitter, unregister_emitter, get_emitter
# emitter type
from .domain import Emitter



from ..domain import Emitter, NodeInfo, NodeKind
from ..util import println, csl, cslq

from collections.abc import Mapping
from itertools import chain


class AgnosticEmitter(Emitter):
    def intro(self) -> int:
        if self.cfg.namespace:
            print(f'namespace {self.cfg.namespace}')
        if self.cfg.root:
            print(f'root {self.cfg.root}')
        print()
        return 0

    def enter_node(
            self, lvl: int,
            parent: NodeInfo | None,
            node: NodeInfo,
            implements: Mapping[str, NodeKind],
            props: Mapping[str, str]):
        if reserved_props := props & self.cfg.common_props.keys():
            raise ValueError(f"reserved property names in '{node.name}': {cslq(reserved_props)}")

        println(
            lvl,
            parent.name + ' >' if parent else '',
            node.kind.value, node.name,
            ': ' + csl(implements) if implements else ''
        )

        for propName, propType in chain(props.items(), self.cfg.common_props.items()):
            println(lvl + 1, f'{propName} : {propType}')



from collections.abc import Iterable


def println(lvl: int, *args, **kwargs):
    print(lvl * 4 * ' ', end='')
    print(*args, **kwargs)


def remove_prefix(prefix: str, s: str) -> str:
    """
    Removes a specified prefix from a string if it exists.

    Args:
        prefix (str): The prefix to remove from the string.
        s (str): The input string to potentially remove the prefix from.

    Returns:
        str: The string with the prefix removed if it starts with the prefix, otherwise the original string.
    """
    return s[len(prefix):] if s.startswith(prefix) else s


def cslq(iterable: Iterable[str]) -> str:
    """Comma Separated List (Quoted)"""
    return csl(f"'{i}'" for i in iterable)


def csl(iterable: Iterable[str]) -> str:
    """Comma Separated List"""
    return ', '.join(iterable)


def sub(pattern: str, num: int, arg: str):
    """
    Substitutes a numbered placeholder in a pattern string with a given argument.

    Args:
        pattern (str): The string containing placeholders in the format $1, $2, etc.
        num (int): The placeholder number to replace.
        arg (str): The replacement string for the specified placeholder.

    Returns:
        str: The pattern with the specified placeholder replaced by the argument.
    """
    return pattern.replace(f'${num}', arg)



from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from enum import Enum
from functools import cache
import re

from .cfg import Config

AstNode = dict[str, 'AstNode | str'] | None
AstUnionNode = dict[str, AstNode]


class NodeKind(Enum):
    Union = 'union'
    Product = 'product'


@dataclass
class NodeInfo:
    name: str
    kind: NodeKind


class Emitter(ABC):
    def __init__(self, cfg: Config):
        self.__cfg = cfg

    @property
    def cfg(self): return self.__cfg

    def intro(self) -> int:
        """
        Echoes the introduction once at the start.

        Returns:
            int: The indentation level to start at.
        """
        return 0

    @abstractmethod
    def enter_node(self, lvl: int,
                   parent: NodeInfo | None,
                   node: NodeInfo,
                   implements: Mapping[str, NodeKind],
                   props: Mapping[str, str]): pass

    def exit_node(self, lvl: int): pass

    def conclusion(self): pass


def is_do_not_touch(s: str) -> bool:
    return s.startswith('=')


def get_dont_touch_me(s: str) -> str:
    return s[1:] if is_do_not_touch(s) else ''


@cache
def pascalize(snake_case: str) -> str:
    """
    Camelizes the string and replaces its first non-underscore, non_dot character by its uppercase equivalent.
    """
    if s := get_dont_touch_me(snake_case):
        return s
    c = camelize(snake_case)
    return re.sub(r'^([_.]*)([^_.])', lambda m: m.group(1) + m.group(2).upper(), c)


@cache
def camelize(snake_case: str) -> str:
    """
    Replace underscores which are not the first character of the string and are followed by a non-underscore character by the uppercase equivalent of that character.
    """
    if s := get_dont_touch_me(snake_case):
        return s
    return re.sub(r'(?<!^)(?:_|(?<=\.))([^_.])', lambda m: m.group()[-1].upper(), snake_case)



from mkast.cfg import Config, Modifier

def test_merge_cfg_simple_fields():
    base = Config(target='python', root='base_root')
    new = Config(target='typescript', root=None)
    merged = base + new
    assert merged.target == 'typescript'
    assert merged.root == 'base_root'

def test_merge_cfg_sets_and_sequences():
    base = Config(known_types={'A', 'B'}, imports=['x', 'y'])
    new = Config(known_types={'B', 'C'}, imports=['z'])
    merged = base + new
    assert set(merged.known_types) == {'A', 'B', 'C'}
    assert list(merged.imports) == ['x', 'y', 'z']

def test_merge_cfg_dicts_and_modifiers():
    base = Config(common_props={'a': '1'}, modifiers={'?': Modifier(type='opt')})
    new = Config(common_props={'b': '2'}, modifiers={'*': Modifier(type='star')})
    merged = base + new
    assert merged.common_props == {'a': '1', 'b': '2'}
    assert set(merged.modifiers.keys()) == {'?', '*'}

def test_merge_cfg_nested_models():
    base = Config(modifiers={'': Modifier(type='base', must='yes')})
    new = Config(modifiers={'': Modifier(type='new', must=None)})
    merged = base + new
    assert merged.modifiers[''].type == 'new'
    assert merged.modifiers[''].must == 'yes'

def test_merge_cfg_none_handling():
    base = Config(root='toor')
    new = Config(root=None)
    merged = base + new
    assert merged.root == 'toor'

def test_merge_cfg_empty_new():
    base = Config(target='python', known_types={'A'})
    new = Config()
    merged = base + new
    assert merged.target == 'python'
    assert set(merged.known_types) == {'A'}


# tests/test_cli.py
from heapq import merge
import io
import argparse as ap
import pathlib
from typing import Any, Mapping
import pydantic
import pytest
from mkast import cfg
import mkast.cli as cli


class FakeConfig(cfg.FileConfig):
    # Minimal stand-in for your real Config for unit tests

    # For resolve_config merge tracking
    origin: str | None = None
    seen: list[str] = pydantic.Field(default_factory=list)

    # Capture what config_from_args fed into model_validate
    last_input: dict | None = None

    def merge(self, new: cfg.Config):
        # Accumulate the origin of each loaded config into `seen`
        assert isinstance(new, FakeConfig)
        return FakeConfig(
            seen=[*self.seen, new.origin] if new.origin else [*self.seen],
            # carry over last extends just to keep loop logic happy
            extends=new.extends,
        )

    @classmethod
    def model_validate(cls, obj: Any,         *,
                       strict: bool | None = None,
                       from_attributes: bool | None = None,
                       context: Any | None = None,
                       by_alias: bool | None = None,
                       by_name: bool | None = None):
        # Record exactly what the helper tried to validate
        cls.last_input = dict(obj)
        # Only keep fields this fake model knows about
        fields = {k: v for k, v in obj.items() if k in cls.model_fields}
        return cls(**fields)


def fn_fake_read_config_file(extends_map: Mapping[str, str | None]):
    def fake_read_config_file(path: pathlib.Path):
        spath = str(pathlib.Path(path).resolve())
        return FakeConfig(extends=extends_map[spath], origin=spath, seen=[spath])
    return fake_read_config_file


def test_prop_ok_basic():
    assert cli.prop("name:type") == ("name", "type")


def test_prop_ok_colon_in_type():
    # Only the first ':' splits
    assert cli.prop("foo:map<string:int>") == ("foo", "map<string:int>")


def test_prop_error_missing_separator():
    with pytest.raises(ap.ArgumentTypeError):
        cli.prop("notValid")


def test_config_from_args_filters_none_and_calls_model_validate(monkeypatch):
    # Patch Config to our FakeConfig
    monkeypatch.setattr(cli, "Config", FakeConfig)

    # Namespace with some None that should be filtered out,
    # plus keys with dashes that your real model likely handles via aliases.
    ns = ap.Namespace(
        target="c",
        **{"known-types": {"X"}, "common-props": [("k", "v")], "ignore_me": None}
    )

    cfg = cli.config_from_args(ns)
    assert isinstance(cfg, FakeConfig)
    # None-valued key should not be passed through
    assert FakeConfig.last_input is not None and "ignore_me" not in FakeConfig.last_input
    # The keys are passed as-is (including dashed names)
    assert FakeConfig.last_input.get("known-types") == {"X"}
    assert FakeConfig.last_input.get("common-props") == [("k", "v")]
    assert cfg.target == "c"


def test_resolve_config_relative_chain(monkeypatch, tmp_path):
    # Layout:
    # tmp/a/config.yaml  -> extends: sub/b.yaml
    # tmp/a/sub/b.yaml   -> extends: ../c.yaml
    # tmp/a/c.yaml       -> extends: null
    a_dir = tmp_path / "a"
    sub_dir = a_dir / "sub"
    sub_dir.mkdir(parents=True)
    file_a = a_dir / "config.yaml"
    file_b = sub_dir / "b.yaml"
    file_c = a_dir / "c.yaml"
    for f in (file_a, file_b, file_c):
        f.write_text("# stub\n", encoding="utf-8")

    extends_map = {
        str(file_a.resolve()): "sub/b.yaml",
        str(file_b.resolve()): "../c.yaml",
        str(file_c.resolve()): None,
    }

    monkeypatch.setattr(cli, "Config", FakeConfig)
    monkeypatch.setattr(cli, "read_config_file", fn_fake_read_config_file(extends_map))

    out = cli.read_config(file_a.resolve())
    assert isinstance(out, FakeConfig)

    assert out.seen == [str(file_a.resolve()), str(file_b.resolve()), str(file_c.resolve())]


def test_resolve_config_cycle_breaks(monkeypatch, tmp_path):
    # A -> B -> A (cycle)
    a_dir = tmp_path / "a"
    a_dir.mkdir(parents=True)
    file_a = a_dir / "config.yaml"
    file_b = a_dir / "b.yaml"
    for f in (file_a, file_b):
        f.write_text("# stub\n", encoding="utf-8")

    extends_map = {
        str(file_a.resolve()): "b.yaml",
        str(file_b.resolve()): "config.yaml",  # cycle
    }

    monkeypatch.setattr(cli, "Config", FakeConfig)
    monkeypatch.setattr(cli, "read_config_file", fn_fake_read_config_file(extends_map))

    out = cli.read_config(file_a.resolve())
    # Should process A then B, then stop due to visited set
    assert isinstance(out, FakeConfig)
    assert out.seen == [str(file_a.resolve()), str(file_b.resolve())]


def test_load_input_single_document(monkeypatch):
    # Patch Config to our FakeConfig so model_validate exists
    monkeypatch.setattr(cli, "Config", FakeConfig)

    data = io.StringIO("root: null\n")
    cfg_from_input, ast = cli.load_input(data)
    assert cfg_from_input is None
    assert ast == {"root": None}


def test_load_input_two_documents(monkeypatch):
    monkeypatch.setattr(cli, "Config", FakeConfig)

    yaml_text = """\
target: c
---
Root:
  child: null
"""
    cfg_from_input, ast = cli.load_input(io.StringIO(yaml_text))
    assert isinstance(cfg_from_input, FakeConfig)
    assert cfg_from_input.target == "c"
    assert ast == {"Root": {"child": None}}


def test_load_input_zero_documents_raises():
    with pytest.raises(ValueError) as ei:
        cli.load_input(io.StringIO(""))
    # Be explicit about the message to catch regressions
    assert "expected 1 or 2 documents, not 0" in str(ei.value)


def test_load_input_three_documents_raises():
    text = """\
---
{}
---
{}
---
{}
"""
    with pytest.raises(ValueError) as ei:
        cli.load_input(io.StringIO(text))
    assert "expected 1 or 2 documents, not 3" in str(ei.value)


def test_load_input_invalid_ast_raises_validation_error(monkeypatch):
    monkeypatch.setattr(cli, "Config", FakeConfig)
    # Second doc is invalid for AstUnionNodeModel (list instead of mapping)
    text = """\
target: c
---
- not
- a
- mapping
"""
    with pytest.raises(pydantic.ValidationError):
        cli.load_input(io.StringIO(text))



from .domain import Config, Emitter, NodeInfo, NodeKind, AstNode, AstUnionNode, is_do_not_touch
from .util import csl, cslq
from .targets.agnostic import AgnosticEmitter
from .targets.csharp import CSharpEmitter

from collections.abc import Iterable, Mapping, Set
from typing import TypeGuard

_emitters: dict[str, type[Emitter]] = {
    'agnostic': AgnosticEmitter,
    'csharp': CSharpEmitter,
}


def get_emitter(cfg: Config) -> Emitter | None:
    cls = _emitters.get(cfg.target, None)
    return None if cls is None else cls(cfg)


def register_emitter(key: str, cls: type[Emitter]) -> type[Emitter]:
    if key in _emitters:
        raise ValueError('emitter already exists at', key, _emitters[key])
    _emitters[key] = cls
    return cls


def unregister_emitter(key: str) -> type[Emitter] | None:
    e = _emitters.get(key)
    if e is not None:
        del _emitters[key]
    return e


def generate_ast(cfg: Config, ast: AstUnionNode, emitter: Emitter | None = None):
    if emitter is None:
        if (emitter := get_emitter(cfg)) is None:
            raise ValueError('no emitter registered for this target language', cfg.target)
    root_node_info = None if cfg.root is None else NodeInfo(cfg.root, NodeKind.Union)
    lvl = emitter.intro()
    for name, node in ast.items():
        walk(emitter, lvl, root_node_info, ast, name, node)
    emitter.conclusion()


def walk(emitter: Emitter,
         lvl: int,
         parent: NodeInfo | None,
         reachable_nodes: AstUnionNode,
         name: str,
         node: AstNode):
    assert reachable_nodes[name] is node, 'invariant: reachable_nodes contains the current node'

    implements = {k: NodeKind.Union for k in in_unions(reachable_nodes, name) if parent is None or k != parent.name}
    if node_is_union(node):
        if redefined_nodes := {k for k in node & reachable_nodes.keys() if node[k] is not None}:
            raise ValueError(f"redefined nodes in '{name}': {cslq(redefined_nodes)}")

        me = NodeInfo(name, NodeKind.Union)
        emitter.enter_node(lvl, parent, me, implements, {})
        for sub in ((k, v) for k, v in node.items() if k not in reachable_nodes.keys()):
            walk(emitter, lvl + 1, me, node | reachable_nodes, *sub)
        emitter.exit_node(lvl)
    else:
        if node is None:
            node = {}

        subs = subnodes(node)
        if redefined_subs := subs & reachable_nodes.keys():
            raise ValueError(f"redefined subnodes in '{name}': {cslq(redefined_subs)}")
        props = {k: v for k, v in node.items() if isinstance(v, str)}
        if undef_type_props := tuple(f"'{k}' ('{v}')" for k, v in props.items() if not check_type(
                emitter.cfg.known_types, reachable_nodes, v)):
            raise ValueError(f"properties of undefined type in '{name}': {csl(undef_type_props)}")

        me = NodeInfo(name, NodeKind.Product)
        emitter.enter_node(lvl, parent, me, implements, props)
        for sub in subs.items():
            walk(emitter, lvl + 1, me, subs | reachable_nodes, *sub)
        emitter.exit_node(lvl)


def check_type(known_types: Set[str], reachable_nodes: AstUnionNode, ptype: str) -> bool:
    real_type = ptype.rstrip('*+?')
    if is_do_not_touch(ptype) or real_type in known_types:
        return True
    s = real_type.split('.', 1)
    if len(s) == 1:
        return s[0] in reachable_nodes.keys()
    first, others = s
    return check_type(
        known_types, reachable_nodes, first) and check_type(
        known_types, reachable_nodes | subnodes(reachable_nodes.get(first, None)),
        others)


def subnodes(node: AstNode) -> AstUnionNode:
    return {} if node is None else {k: v for k, v in node.items() if not isinstance(v, str)}


def in_unions(reachable_nodes: Mapping[str, AstNode], name: str) -> Iterable[str]:
    """ Returns the names of each union this node is in"""
    for k, v in reachable_nodes.items():
        if node_is_union(v):
            if name in v:
                yield k
        else:
            yield from in_unions(subnodes(v), name)


def node_is_union(node: AstNode) -> TypeGuard[AstUnionNode]:
    return node is not None and not any(isinstance(v, str) for v in node.values())
